// lptdos.pas - parallel port server for Linux
// Compile with
// ppc386 -S2 -Rintel lptdos.pas
// su
// chown root lptdos
// chmod +s lptdos
//
// run with:
// ./lptdos from USER mode
// will use seteuid(getuid) after allocating port

program lptdos;

uses dos,linux,crt;

{$i-}

type blockarr=array[0..3] of byte;

const blocksz=4096;
const blocklast=4099;
type pdatabuftype=^databuftype;
     databuftype=record
       blockheader:blockarr;              {Will be set by called function!}
       blocksize,
       blocknr:longint;
       blockdata:array[0..blocklast] of char;  {Inkl. CRC!}
     end;

const blockheader:blockarr=($18,$24,$42,$81);   {Test all bits!}

const
  parerr_ok=0;
  parerr_file=1;
  parerr_dir=2;
  parerr_aborted=3;
  parerr_timeout=4;
  parerr_fatal=5;
  parerr_inuse=6;

const
  INVALID_HANDLE_VALUE=0;

var blocktransfer:boolean;             {currently in a block transfer?}
    abortclicked:boolean;              {Checks for any key}
    portaddr,portaddr1:word;

procedure sleep(ms:integer);
begin
  delay(ms);
end;

function ReadPort(p:word):byte;
begin
  asm
    mov dx,p
    in al,dx
    mov result,al
  end;
end;

procedure WritePort(p:word;b:byte);
begin
  asm
    mov dx,p
    mov al,b
    out dx,al
  end;
end;

function GetPortAddress(lptportnr:integer):integer; { 1 for LPT1, 2 for LPT2, 3 for LPT3 }
(*var f:file;
    dataread:integer;
    portnr:word;*)
begin
(*   // unfortunately, /dev/mem doesn't seem to access the BIOS numbers...
  assign(f,'/dev/mem');
  filemode:=0;
  reset(f,1);
  if ioresult<>0 then exit;
  seek(f,$00408+2*(lptportnr));
  blockread(f,@portnr,2, dataread);
  close(f);
  writeln('Portnr from /dev/mem: ',portnr);
*)
  case lptportnr of
    1:result:=$378; {888} //MemW[$0040:$0008+2*(lptportnr-1)];
    2:result:=$3BC; {956}
  else
    result:=$278;   {632}
  end;
  if not IOperm(result,3,6) then
    result:=0;
end;

{************************ Low level functions *************************}

function getstatusbits:byte;
begin
  getstatusbits:=ReadPort(portaddr+1);
end;

function WaitForBitFlip(bithigh:boolean):boolean;
var starttime:longint;
    wait:word;
    testbit:boolean;
begin
  wait:=10000;
  repeat
    testbit:=ReadPort(portaddr+1) and $80<>0;
    dec(wait)
  until (wait=0) or (bithigh=testbit);
  if bithigh<>testbit then begin     {No reaction so far! -> try for up to 10 seconds}
    starttime:=20;
    repeat
      wait:=10000;
      repeat
        testbit:=ReadPort(portaddr+1) and $80<>0;
        dec(wait)
      until (wait=0) or (bithigh=testbit);
      sleep(100);
      dec(starttime);
    until (wait<>0) or (starttime<=0); 
  end;
  WaitForBitFlip:=wait<>0;
end;

procedure SetPortReady;
begin
  WritePort(portaddr,0);  {Ready}
end;

procedure SendDisableInterrupts;
begin
  WritePort(portaddr+2,4);  {disable interrupts}
end;

procedure SignalConnectionError;
begin
  WritePort(portaddr,$E);
  sleep(50);
  WritePort(portaddr,$1E);
  sleep(50);
  WritePort(portaddr,$E);
  sleep(50);
  WritePort(portaddr,$1E);   {Signal a connection error}
  sleep(50);
end;

procedure SignalAbortTransfer;
begin
  WritePort(portaddr,$1E);  {Abort transfer}
  sleep(100);
  WritePort(portaddr,$0E);  {Abort transfer}
  sleep(100);
end;

type chararray=array[0..5000] of char;
type pchar=^chararray;

function inttostr(i:longint):string;
var s:string;
begin
  str(i,s);
  inttostr:=s;
end;

function strtoint(s:string):longint;
var i:longint;
    code:integer;
begin
  val(s,i,code);
  if code<>0 then strtoint:=-1
             else strtoint:=i;
end;

{************************ Send/receive data *************************}

function UploadBlock(var data:databuftype):integer;
var i,tst,size,sizem4:integer;
    crc32:longint;
    crc32_1,crc32_2:word;
    pdata,pcrc:pchar;
    thischar:char;
    ackreceived:boolean;
    servererror:integer;
begin
  SendDisableInterrupts;
  servererror:=0;
  ackreceived:=true;
  crc32:=-1;
  crc32_1:=1;
  crc32_2:=0;
  size:=data.blocksize+16;   {Inkl. Header und CRC!}
  data.blockheader:=blockheader;
  pdata:=pchar(@data);
  pcrc:=pchar(@crc32);
  i:=0;
  ackreceived:=true;
  sizem4:=size-4;

  while (i<12) and ackreceived do begin   {Blockaufbau: 4 byte header, 4 byte Blockgroesse, 4 byte blocknr,
                                             Blockdaten, 4 byte crc}
    thischar:=pdata^[i];
    WritePort(portaddr,16 or (ord(thischar) and 15)); {Low nibble}
    ackreceived:=WaitForBitFlip(false);
    if ackreceived then begin
      WritePort(portaddr,ord(thischar) div 16); {High nibble}
      ackreceived:=WaitForBitFlip(true);
    end;
    if ackreceived then begin
      tst:=ReadPort(portaddr1) and $78; {Only data part!}
      if tst<>0 then begin
        if (tst=$78) then begin    {Re-send block!}
          i:=-1;
          crc32_1:=1;
          crc32_2:=0;
          crc32:=-1;
        end else if (tst=$70) then begin  {Transmission aborted!}
          ackreceived:=false;
          servererror:=parerr_fatal;
        end else if (tst=$68)  then begin  {Transmission aborted!}
          ackreceived:=false;
          servererror:=parerr_file;
        end;
      end;
      inc(i);
    end;
  end;

  while (i<sizem4) and ackreceived do begin   {Blockaufbau: 4 byte header, 4 byte Blockgroesse, 4 byte blocknr,
                                             Blockdaten, 4 byte crc}
    thischar:=pdata^[i];
    inc(crc32_1,byte(thischar));
    inc(crc32_2,crc32_1);
    WritePort(portaddr,16 or (ord(thischar) and 15)); {Low nibble}
    ackreceived:=ReadPort(portaddr1) and $80=0;
    if not ackreceived then
      ackreceived:=ReadPort(portaddr1) and $80=0;
    if not ackreceived then
      ackreceived:=WaitForBitFlip(false);
    if ackreceived then begin
      WritePort(portaddr,ord(thischar) div 16); {High nibble}
      ackreceived:=ReadPort(portaddr1) and $80<>0;
      if not ackreceived then
        ackreceived:=ReadPort(portaddr1) and $80<>0;
      if not ackreceived then
        ackreceived:=WaitForBitFlip(true);
    end;
    if ackreceived then begin
      tst:=ReadPort(portaddr1) and $78; {Only data part!}
      if tst<>0 then begin
        if (tst=$78) then begin    {Re-send block!}
          i:=-1;
          crc32_1:=1;
          crc32_2:=0;
          crc32:=-1;
        end else if (tst=$70) then begin  {Transmission aborted!}
          ackreceived:=false;
          servererror:=parerr_fatal;
        end else if (tst=$68)  then begin  {Transmission aborted!}
          ackreceived:=false;
          servererror:=parerr_file;
        end;
      end;
      inc(i);
    end;
  end;
  crc32:=65536*crc32_2+crc32_1;

  while (i<size) and ackreceived do begin   {Blockaufbau: 4 byte header, 4 byte Blockgroesse, 4 byte blocknr,
                                             Blockdaten, 4 byte crc}
    thischar:=pcrc^[i-size+4];
    WritePort(portaddr,16 or (ord(thischar) and 15)); {Low nibble}
    ackreceived:=WaitForBitFlip(false);
    if ackreceived then begin
      WritePort(portaddr,ord(thischar) div 16); {High nibble}
      ackreceived:=WaitForBitFlip(true);
    end;
    if ackreceived then begin
      tst:=ReadPort(portaddr1) and $78; {Only data part!}
      if tst<>0 then begin
        if (tst=$78) then begin    {Re-send block!}
          i:=-1;
          crc32_1:=1;
          crc32_2:=0;
          crc32:=-1;
        end else if (tst=$70) then begin  {Transmission aborted!}
          ackreceived:=false;
          servererror:=parerr_fatal;
        end else if (tst=$68)  then begin  {Transmission aborted!}
          ackreceived:=false;
          servererror:=parerr_file;
        end;
      end;
      inc(i);
    end;
  end;
  if not ackreceived then begin
    WritePort(portaddr,0);  {Ready}
    sleep(50);
  end;
  if servererror<>0 then
    UploadBlock:=servererror
  else if ackreceived then
    UploadBlock:=parerr_ok
  else
    UploadBlock:=parerr_timeout;
end;

function bytetohex(b:byte):string;
const hexs:array[0..15] of char='0123456789ABCDEF';
begin
  bytetohex:=hexs[b shr 4]+hexs[b and 15];
end;

type plongint=^longint;

function DownloadBlock(var data:databuftype):integer;
var crc32,blocksz,sentcrc,sentblocknr:longint;
    crc32_1,crc32_2:word;
    recvblockheader:blockarr;
    dataidx,retransmits:integer;
    rawsz,rawnr,rawdata:pchar;
    lonibble,hinibble:byte;
    receiving:boolean;
    thischar:char;
begin
  SendDisableInterrupts;
  retransmits:=0;
  dataidx:=0;
  blocksz:=0;
  crc32:=-1;
  crc32_1:=1;
  crc32_2:=0;
  result:=parerr_ok;
  rawdata:=pchar(@data);
  rawsz:=pchar(@blocksz);
  rawnr:=pchar(@sentblocknr);
  repeat
    receiving:=ReadPort(portaddr1) and $80=0;
    if not receiving then
      receiving:=ReadPort(portaddr1) and $80=0;
    if not receiving then
      receiving:=WaitForBitFlip(false);

    if receiving then begin
      lonibble:=ReadPort(portaddr1);    {Nochmal lesen <- bit-Kippfehler}
      WritePort(portaddr,16);  {Ack}

      receiving:=ReadPort(portaddr1) and $80<>0;
      if not receiving then
        receiving:=ReadPort(portaddr1) and $80<>0;
      if not receiving then
        receiving:=WaitForBitFlip(true);

      if receiving then begin
        hinibble:=ReadPort(portaddr1);    {Nochmal lesen <- bit-Kippfehler}
        thischar:=char((hinibble shl 1) and $F0+(lonibble shr 3) and $F);
        if dataidx>=12 then begin
          rawdata^[dataidx]:=thischar;
          if (dataidx-12<blocksz) then begin
            inc(crc32_1,byte(thischar));
            inc(crc32_2,crc32_1);
          end;
        end else if dataidx<4 then
          recvblockheader[dataidx]:=byte(thischar)
        else if dataidx<8 then begin
          rawsz^[dataidx-4]:=thischar;
          if (dataidx=4) and not (longint(recvblockheader)=longint(blockheader)) then begin  {Fatal error: bad block!}
            {Check if by 1 byte shifted!}
            if (recvblockheader[1]=blockheader[0]) and
               (recvblockheader[2]=blockheader[1]) and
               (recvblockheader[3]=blockheader[2]) then begin
              dec(dataidx);
              recvblockheader:=blockheader;
            end else if (recvblockheader[0]=blockheader[1]) and
                        (recvblockheader[1]=blockheader[2]) and
                        (recvblockheader[2]=blockheader[3]) then begin
              inc(dataidx);
              rawsz^[0]:=char(recvblockheader[3]);
              rawsz^[1]:=thischar;
              recvblockheader:=blockheader;
            end else begin
              result:=parerr_fatal;
              receiving:=false;
            end;
          end;
        end else begin
          rawnr^[dataidx-8]:=thischar;
        end;
        inc(dataidx);
        if (dataidx=blocksz+16) then begin {flush!}
          crc32:=65536*crc32_2+crc32_1;
          sentcrc:=plongint(@data.blockdata[dataidx-16])^;
          {Store it here!}
          if sentblocknr=-1 then begin  {Abort sign}
            WritePort(portaddr,13);  {Abort transfer, no retry}
            result:=parerr_aborted;
          end else if (sentcrc<>crc32) then begin  {Wrong crc -> try again!}
            inc(retransmits);
            WritePort(portaddr,15);  {Nack - data bad - causes block resend}
            if retransmits>10 then begin
              result:=parerr_fatal;
            end;
          end else if sentblocknr<>data.blocknr then begin  {Wrong block -> fatal error}
            if (data.blocknr=-1) then
              WritePort(portaddr,13)   {Abort transfer, no retry}
            else
              WritePort(portaddr,14);  {Abort transfer}
            if (data.blocknr=-1) or (abs(sentblocknr-data.blocknr)<=2) then
              result:=parerr_aborted      {Error at block end!}
            else
              result:=parerr_fatal;
          end else begin
            WritePort(portaddr,0);  {Ack}
            data.blocksize:=blocksz;
            retransmits:=0;
            receiving:=false;       {Success!}
          end;
          crc32_1:=1;
          crc32_2:=0;
          crc32:=-1;
          dataidx:=0;
        end else
          WritePort(portaddr,0);  {Ack}
      end else result:=parerr_fatal;
    end else result:=parerr_fatal;
  until not receiving or (result<>parerr_ok);
  if result=parerr_fatal then begin
    SignalAbortTransfer;       {Abort transfer}
    SetPortReady;              {Clear error condition}
  end;
  DownloadBlock:=result;
end;





{************************ Send/receive commands *************************}

function RemoteSendCommand(command:string):boolean;
var data:databuftype;
    i:integer;
begin
  if blocktransfer then begin
    RemoteSendCommand:=false;
    exit
  end;
  blocktransfer:=true;
  for i:=1 to length(command) do
    data.blockdata[i-1]:=command[i];
  data.blockdata[length(command)]:=char(0);
  data.blocksize:=length(command);
  data.blocknr:=0;
  RemoteSendCommand:=UploadBlock(data)=parerr_ok;
  blocktransfer:=false;
end;

function ReceiveCommandString(var s:string):shortint;
var databuf:databuftype;
    i:integer;
begin
  if blocktransfer then begin
    ReceiveCommandString:=parerr_inuse;
    exit
  end;
  blocktransfer:=true;
  databuf.blocknr:=0;
  if DownloadBlock(databuf)=0 then begin
    s:='';
    for i:=0 to databuf.blocksize-1 do
      s:=s+databuf.blockdata[i];
    ReceiveCommandString:=1
  end else
    ReceiveCommandString:=-1;
  blocktransfer:=false;
end;

{************************ Receive file (client or server!) *************************}

function ReceiveData(filename:string;filesize,filetime,fileattr,RestartAt,RestartAtRemote:longint;Asciimode:boolean):integer;
var data:databuftype;
    f:file;
    compdata:array[0..511] of char;
    asciidata:array[0..blocksz] of char;
    totalreceived,blocknr:longint;
    err,ioerr,i,AsciiIndex:integer;
    written:word;
    receiving,compareOK:boolean;
    ch:char;
begin
  if blocktransfer then begin
    ReceiveData:=parerr_inuse;
    exit
  end;
  ReceiveData:=parerr_ok;
  blocktransfer:=true;
  if filesize-RestartAt=0 then begin
    if RestartAt=0 then begin
      assign(f,filename);
      rewrite(f,1);                        {Create empty file!}
      if ioresult=0 then begin
        truncate(f);
        if filetime<>-1 then
          setftime(f,filetime);
        close(f);
        if fileattr<>-1 then
          setfattr(filename,fileattr);
        ReceiveData:=parerr_ok;
      end else
        ReceiveData:=parerr_file;
    end else
      ReceiveData:=parerr_ok;
    blocktransfer:=false;
    exit;
  end;
  totalreceived:=RestartAtRemote;
  blocknr:=1;
  repeat
    data.blocknr:=blocknr;
    err:=DownloadBlock(data);
    if err=0 then begin
      receiving:=true;
      if blocknr=1 then begin          {Data OK -> Open local file!}
        assign(f,filename);
        if RestartAt=0 then begin
          rewrite(f,1);
          ioerr:=ioresult;
        end else begin
          filemode:=2;
          reset(f,1);                  {Open existing for reading/writing!}
          ioerr:=ioresult;
          if ioerr=0 then begin
            seek(f,RestartAt);
            ioerr:=ioresult;
            if ioerr=0 then begin
              compareOK:=true;
              blockread(f,compdata[0],512,written);
              ioerr:=ioresult;
              if (ioerr=0) and (written=512) then begin
                if AsciiMode then begin
                  asciiindex:=0;
		  for i:=0 to written-1 do begin
		    ch:=data.blockdata[i];
                    if ch<>#13 then begin
		      if compdata[asciiIndex]<>ch then compareOK:=false;
		      if not compareOK then
		        write(i,':',ch,'-',compdata[asciiIndex],'   ');
		      inc(AsciiIndex);
		    end;
                  end;
		end else
                  for i:=0 to written-1 do
                    if compdata[i]<>data.blockdata[i] then compareOK:=false;
              end else doserror:=0;         {Append: nothing to compare!}
              if not compareOK then begin
                writeln('Error, source and target different!');
                blocknr:=-1;
                abortclicked:=true;
              end else begin
                seek(f,RestartAt);
                ioerr:=ioresult;
              end;
            end; {ioerr=0}
          end; {ioerr=0}
        end;  {not restartat}
        if ioerr<>0 then               {Send abort signal!}
          blocknr:=-1;                 {Send blocknr=-1 as abort signal!}
      end;    {Blocknr=1}              {Will abort in next block!}
      if blocknr<>-1 then begin
        inc(blocknr);
        inc(totalreceived,data.blocksize);
	if AsciiMode then begin
          asciiindex:=0;
	  for i:=0 to data.blocksize-1 do begin
	    ch:=data.blockdata[i];
	    if ch<>#13 then begin
	      asciidata[AsciiIndex]:=ch;
	      inc(AsciiIndex);
	    end;
	  end;    
	  blockwrite(f,asciidata,asciiindex,written);
	end else
          blockwrite(f,data.blockdata,data.blocksize,written);
        if blocknr mod 25=0 then
          write('*');
        ReceiveData:=parerr_ok;
      end;
    end else begin
      writeln(totalreceived,' ERROR');
      receiving:=false;
      ReceiveData:=err;
    end;
  until not receiving or (totalreceived=filesize);
  if filetime<>-1 then
    setftime(f,filetime);
  close(f);
  if fileattr<>-1 then
    setfattr(filename,fileattr);

  blocktransfer:=false;
end;

{******************************** Server functions ******************************}

function ServerListDirectory(dirname:string):boolean;
var databuf:databuftype;
    inbuf:string;
    totaldatasize,starttimeseconds:longint;
    totalsent,blocknr:longint;
    l,blocktotal,i:integer;
    s:searchrec;
    dataread:word;
    ok:boolean;
begin
  if blocktransfer then begin
    ServerListDirectory:=false;
    exit
  end;
  starttimeseconds:=GetEpochTime;

  if dirname='' then dirname:='\';
  for i:=1 to length(dirname) do
    if dirname[i]='\' then dirname[i]:='/';
  if dirname[1]<>'/' then dirname:='/'+dirname;
  
  begin
    if length(dirname)>=1 then begin
      if dirname[length(dirname)]='/' then
        dirname:=dirname+'*.*';
    end;

    FindFirst(dirname,anyfile,s);
    if doserror<>0 then begin
      ServerListDirectory:=RemoteSendCommand('DIR ERROR '+inttostr(doserror));
      exit;
    end;

    totaldatasize:=0;
    ok:=true;

    while (doserror=0) and ok do begin
      if (s.name<>'..') and (s.name<>'.') then begin
        inbuf:=s.name+#9+inttostr(s.size)+#9+inttostr(s.time)+#9+inttostr(s.attr)+#13#10;
        inc(totaldatasize,length(inbuf));
  //      write(s.name,' ');
      end;
      if (GetEpochTime-starttimeseconds>=1) then begin
        starttimeseconds:=GetEpochTime;
	ok:=RemoteSendCommand('WAIT');
      end;
      FindNext(s);
    end; {while doserror=0}
    FindClose(s);
  end;

  ok:=RemoteSendCommand('SIZE '+inttostr(totaldatasize));
  if ok and (totaldatasize>0) then begin  {Only send data if file<>0!!!}
    begin
      FindFirst(dirname,anyfile,s);
      totalsent:=0;
      dataread:=1;
      blocknr:=1;
      ok:=true;
      abortclicked:=false;
      blocktransfer:=true;
      blocktotal:=0;
      while (doserror=0) do begin
        if (s.name<>'..') and (s.name<>'.') then begin
          inbuf:=s.name+#9+inttostr(s.size)+#9+inttostr(s.time)+#9+inttostr(s.attr)+#13#10;

          l:=length(inbuf);
          if ok and (l>0) then begin
            move(inbuf[1],databuf.blockdata[blocktotal],l);
            inc(blocktotal,l);
            if blocktotal>blocksz-200 then begin  {Flush!}
              abortclicked:=keypressed;
              if abortclicked then begin
                databuf.blocksize:=0;
                databuf.blocknr:=-1;
                ok:=UploadBlock(databuf)=parerr_ok; {Signal abort!}
                ok:=false;
              end else begin
                databuf.blocknr:=blocknr;
                databuf.blocksize:=blocktotal;
                ok:=UploadBlock(databuf)=parerr_ok;
                inc(totalsent,blocktotal);
                inc(blocknr);
                write('*');
              end;
              blocktotal:=0;
            end;
          end;
        end;
        FindNext(s);
      end; {while doserror=0}
      FindClose(s);
    end; {not getdrivelist}
    if (blocktotal>0) and ok then begin
      blocktransfer:=true;
      databuf.blocknr:=blocknr;
      databuf.blocksize:=blocktotal;
      ok:=UploadBlock(databuf)=parerr_ok;
      inc(totalsent,blocktotal);
    end;
    blocktransfer:=false;
  end;
  ServerListDirectory:=ok;
end;

function min(a,b:integer):integer;
begin
  if a<b then result:=a
         else result:=b; 
end;

function ServerUploadFile(filename:string;RestartAt:longint;AsciiMode:boolean):boolean;
const readblocksz=32768;
var f:file;
    readblock:array[0..readblocksz-1] of char;
    pdatabuf:pdatabuftype;
    totalsent,blocknr,ftime,thesize,thesizedisk,restartatdisk,starttimeseconds:longint;
    i,blocki:integer;
    dataread,fattr:word;
    ok,nextisnr13,onlyleadinglf:boolean;
    ch:char;
begin
  if blocktransfer then begin
    ServerUploadFile:=false;
    exit
  end;
  filemode:=0;
  assign(f,filename);
  reset(f,1);
  if (ioresult=0) and (RestartAt<>0) then
    Seek(f,RestartAt);
  if ioresult<>0 then begin
    ServerUploadFile:=RemoteSendCommand('FILE ERROR '+inttostr(doserror));
    exit;
  end;
  ok:=true;
  if RestartAt<>0 then
    ok:=RemoteSendCommand('REST '+inttostr(RestartAt));
  if ok then begin
    GetFTime(f,ftime);
    ok:=RemoteSendCommand('TIME '+inttostr(ftime));
  end;
  if ok then begin
    GetFAttr(f,fattr);
    ok:=RemoteSendCommand('ATTR '+inttostr(fattr));
  end;
  onlyleadinglf:=false;
  if AsciiMode then begin
    thesize:=0;
    thesizedisk:=0;
    restartatdisk:=0;
    dataread:=1;
    Seek(f,0);
    starttimeseconds:=GetEpochTime;
    while (dataread>0) and ok do begin
      blockread(f,readblock[0],readblocksz,dataread);
      if Restartat=0 then begin
        inc(thesize,dataread);
        for i:=0 to dataread-1 do
          if readblock[i]=#10 then inc(thesize);
      end else begin    (* RestartAt position is counted including CRLF *)
        for i:=0 to dataread-1 do begin
          inc(thesize);
	  inc(thesizedisk);
	  if readblock[i]=#10 then inc(thesize);
	  if (RestartAtDisk=0) and (thesize>=RestartAt) then begin
	    RestartAtDisk:=thesizedisk;
	    onlyleadinglf:=thesize>restartat;	  
	  end;
	end;  
        if (GetEpochTime-starttimeseconds>=1) then begin
          starttimeseconds:=GetEpochTime;
	  ok:=RemoteSendCommand('WAIT');
        end;	
      end;	  
    end;
    seek(f,RestartAtDisk);
  end else
    thesize:=filesize(f);
  
  if ok then ok:=RemoteSendCommand('SIZE '+inttostr(thesize));
  if ok then begin
    totalsent:=RestartAt;
    dataread:=1;
    blocknr:=1;
    abortclicked:=false;
    blocktransfer:=true;
    new(pdatabuf);
    while (dataread>0) and ok do begin
      blockread(f,readblock[0],readblocksz,dataread);
      if (ioresult=0) and (dataread>0) then begin
        blocki:=0;
	i:=0;
	nextisnr13:=false;
	while ok and ((i<dataread) or nextisnr13) do begin
	  if AsciiMode then begin
  	    if nextisnr13 then begin
	      pdatabuf^.blockdata[blocki]:=#10;
	      nextisnr13:=false;
	    end else begin
	      ch:=readblock[i];
	      nextisnr13:=ch=#10;
	      if nextisnr13 then ch:=#13;
	      if onlyleadinglf and nextisnr13 then begin
	        dec(blocki);
		onlyleadinglf:=false;
	      end else begin
                pdatabuf^.blockdata[blocki]:=ch;
	        inc(i);
	      end;
	    end;  
	    inc(blocki);
	  end else begin
	    blocki:=min(blocksz,dataread-i);
	    move(readblock[i],pdatabuf^.blockdata[0],blocki);
	    inc(i,blocki);
	  end;  
	  if (blocki=blocksz) or ((i=dataread) and not nextisnr13) then begin
	    pdatabuf^.blocknr:=blocknr;
            pdatabuf^.blocksize:=blocki;
            ok:=UploadBlock(pdatabuf^)=parerr_ok;
            if blocknr mod 25=0 then
              write('*');
            inc(totalsent,blocki);
	    blocki:=0;
            inc(blocknr);
            abortclicked:=keypressed;
            if abortclicked then
              ok:=false;
          end;
	end;   
      end;	
    end;
    dispose(pdatabuf);
    blocktransfer:=false;
  end;
  close(f);
  ServerUploadFile:=ok;
end;


function ServerCreateDirectory(dirname:string):boolean;
var err:integer;
    res:boolean;
begin
  mkdir(dirname);
  err:=ioresult;
  if err=0 then result:=true
  else begin
    chdir(dirname);
    result:=ioresult=0;
  end;
  if result=true then
    res:=RemoteSendCommand('OK')
 else
    res:=RemoteSendCommand('DIR ERROR '+inttostr(err));
end;

function ServerDeleteFile(filename:string):boolean;
var f:file;
    err:integer;
    res:boolean;
begin
  assign(f,filename);
  erase(f);
  err:=ioresult;
  if err<>0 then begin
    setfattr(filename,archive);
    erase(f);
    err:=ioresult;
  end;
  result:=err=0;
  if result then
    res:=RemoteSendCommand('OK')
  else
    res:=RemoteSendCommand('FILE ERROR '+inttostr(doserror));
end;

function ServerRemoveEmptyDirectory(dirname:string):boolean;
var res:boolean;
begin
  rmdir(dirname);
  result:=ioresult=0;
  if result then
    res:=RemoteSendCommand('OK')
  else
    res:=RemoteSendCommand('DIR ERROR '+inttostr(doserror));
end;

function ServerRenameFile(fromname,toname:string):boolean;
var f:file;
    res:boolean;
begin
  Assign(f,fromname);
  Rename(f,toname);
  result:=ioresult=0;
  if result then
    res:=RemoteSendCommand('OK')
  else
    res:=RemoteSendCommand('RENAME ERROR '+inttostr(doserror));
end;

function ServerSignalConnectOK:boolean;
begin
  ServerSignalConnectOK:=RemoteSendCommand('OK A');
end;

function formatunixname(s:string):string;
var i:integer;
begin
  for i:=1 to length(s) do
    if s[i]='\' then s[i]:='/';
  if s[1]<>'/' then s:='/'+s;
  result:=s;
end;

function GetRestartAt(var f:file;RestartAt:longint):longint;
const readblocksz=32768;
var readblock:array[0..readblocksz-1] of char;
    dataread,datatotal,remotedatatotal,i,starttimeseconds:longint;
begin
  result:=0;
  dataread:=1;
  datatotal:=0;
  remotedatatotal:=0;
  starttimeseconds:=GetEpochTime;
  while (dataread>0) and (result=0) do begin
    blockread(f,readblock[0],readblocksz,dataread);
    for i:=0 to dataread-1 do begin
      inc(remotedatatotal);
      inc(datatotal);
      if readblock[i]=#10 then inc(remotedatatotal);
      if datatotal=RestartAt then
        result:=remotedatatotal;
    end;  
    if (GetEpochTime-starttimeseconds>=1) then begin
      starttimeseconds:=GetEpochTime;
      if not RemoteSendCommand('WAIT') then
        result:=-1;
    end;	
  end;    
end;

procedure DoServer;

  procedure addtomemo(s:string);
  begin
    writeln(s);
  end;

  function waitforstatusbit:boolean;
  var statusnr:byte;
      signaled:boolean;
  begin
    repeat
      statusnr:=getstatusbits;  {Call to driver on NT!}
      signaled:=statusnr and $80=0;
      if not signaled then begin
        sleep(50);
      end;
      abortclicked:=keypressed;
    until abortclicked or signaled;
    waitforstatusbit:=signaled and not abortclicked;
  end;


var s,recname,renamedfrom:string;
    rectime,RestartAt,RestartAtRemote:longint;
    recattr:integer;
    recstatus:shortint;   {0=OK, 1=Done, -1=Error}
    f:file;
    actionok,dummy,asciiput:boolean;
begin
  SendDisableInterrupts;
  SetPortReady;
  recstatus:=0;
  recname:='';
  renamedfrom:='';
  rectime:=-1;
  recattr:=-1;
  RestartAt:=0;
  RestartAtRemote:=0;
  abortclicked:=false;
  asciiput:=false;
  while not abortclicked do begin
    s:='';
    if waitforstatusbit then begin  {Wait for wake-up signal!}
      recstatus:=ReceiveCommandString(s);
      if not abortclicked then begin
        if s<>'' then addtomemo(s);
        if copy(s,1,4)='PUT ' then begin
          recname:=copy(s,5,300);
          recname:=formatunixname(recname);
          rectime:=-1;
          recattr:=-1;
	  asciiput:=false;
        end else if copy(s,1,5)='APUT ' then begin
          recname:=copy(s,6,300);
          recname:=formatunixname(recname);
          rectime:=-1;
          recattr:=-1;
	  asciiput:=true;
        end else if copy(s,1,5)='TIME ' then begin
          rectime:=strtoint(copy(s,6,300));
        end else if copy(s,1,5)='ATTR ' then begin
          recattr:=strtoint(copy(s,6,300));
        end else if copy(s,1,6)='REST ?' then begin
          actionok:=false;
          if RecName<>'' then begin
            assign(f,recname);
            reset(f,1);
            if ioresult=0 then begin
              RestartAt:=filesize(f);
              if ioresult<>0 then
                RestartAt:=0;
              RestartAt:=((RestartAt-4096) div 1024) * 1024;
              if RestartAt<0 then
                RestartAt:=0;
	      if asciiput and (RestartAt<>0) then begin
	        RestartAtRemote:=GetRestartAt(f,RestartAt);
	      end else
	        RestartAtRemote:=RestartAt;
              close(f);
	      if RestartAt>=0 then begin
  	        actionok:=true;
                dummy:=RemoteSendCommand('REST='+inttostr(RestartAtRemote));
	      end else begin
	        RestartAtRemote:=0;
		RestartAt:=0;
	      end;
            end;
          end;
          if not actionok then
            dummy:=RemoteSendCommand('REST ERROR '+inttostr(doserror));
        end else if copy(s,1,5)='REST ' then begin
          RestartAt:=strtoint(copy(s,6,300));
          if RestartAt=-1 then
            RestartAt:=0;
        end else if copy(s,1,5)='SIZE ' then begin
	  if ReceiveData(recname,strtoint(copy(s,6,300)),rectime,recattr,RestartAt,RestartAtRemote,asciiput)=0 then begin
            addtomemo('UPLOAD OK');
            recstatus:=1
          end else begin
            addtomemo('UPLOAD FAILED');
            recstatus:=-1;
          end;
	  asciiput:=false;
          abortclicked:=false;
          recname:='';
          rectime:=-1;
          recattr:=-1;
          RestartAt:=0;
        end else if copy(s,1,4)='GET ' then begin
          if ServerUploadFile(formatunixname(copy(s,5,300)),RestartAt,false)
          then begin
            addtomemo('DOWNLOAD OK');
            recstatus:=1
          end else begin
            addtomemo('DOWNLOAD FAILED');
            recstatus:=-1;
          end;
          abortclicked:=false;
          RestartAt:=0;
	  RestartAtRemote:=0;
        end else if copy(s,1,5)='AGET ' then begin
          if ServerUploadFile(formatunixname(copy(s,6,300)),RestartAt,true)
          then begin
            addtomemo('DOWNLOAD OK');
            recstatus:=1
          end else begin
            addtomemo('DOWNLOAD FAILED');
            recstatus:=-1;
          end;
          abortclicked:=false;
          RestartAt:=0;
	  RestartAtRemote:=0;
        end else if copy(s,1,5)='LIST ' then begin
          asciiput:=false;
          if ServerListDirectory(copy(s,6,300)) then begin
            addtomemo('LIST OK');
            recstatus:=1
          end else begin
            addtomemo('LIST FAILED');
            recstatus:=-1;
          end;
          abortclicked:=false;
        end else if copy(s,1,4)='DEL ' then begin
          if ServerDeleteFile(formatunixname(copy(s,5,300))) then
            recstatus:=1
          else
            recstatus:=-1;
        end else if copy(s,1,7)='CONNECT' then begin
          if ServerSignalConnectOK then begin
            writeln('Connection established!');
            recstatus:=1
          end else
            recstatus:=-1;
        end else if copy(s,1,3)='MD ' then begin
          if ServerCreateDirectory(formatunixname(copy(s,4,300))) then
            recstatus:=1
          else
            recstatus:=-1;
        end else if copy(s,1,3)='RD ' then begin
          if ServerRemoveEmptyDirectory(formatunixname(copy(s,4,300))) then
            recstatus:=1
          else
            recstatus:=-1;
        end else if copy(s,1,5)='RNFR ' then begin
          renamedfrom:=formatunixname(copy(s,6,300));
        end else if copy(s,1,5)='RNTO ' then begin
          recstatus:=-1;
          if renamedfrom<>'' then
            if ServerRenameFile(renamedfrom,formatunixname(copy(s,6,300))) then
              recstatus:=1;
        end else begin               {Unknown command}
          SignalConnectionError;     {Signal a connection error}
          SetPortReady;              {Clear error condition}
        end;
      end;
      s:='';
    end;
    if recstatus=-1 then
      SetPortReady;                  {Ready}
  end;
end;

function seteuid(newuid:longint):longint; external 'libc';

procedure StartServer(portidx:integer);
begin
  writeln('Current effective user ID: ',geteuid,
          ', ID of calling user: ',getuid);
  if getuid=geteuid then begin  
    writeln('WARNING: You should run this program as a normal user with');
    writeln('suid root. This is done with the following commands:');
    writeln;
    writeln('su');
    writeln('chown root lptdos');
    writeln('chmod +s lptdos');
    writeln('logout');
    writeln;
  end;

  write('Trying to allocate port: ');
  portaddr:=GetPortAddress(portidx);
  portaddr1:=portaddr+1;
  if portaddr<>0 then writeln('Success!')
                 else writeln('Failed!');
  if getuid<>geteuid then begin
    seteuid(getuid);         // reduce UID to user UID!
    writeln('Current effective user ID: ',geteuid,
            ', ID of calling user: ',Geteuid);
  end;
  if portaddr=0 then exit;

  write('Accepting connections through port LPT',portidx);
  writeln(', port addr: ',portaddr);
  abortclicked:=false;
  blocktransfer:=false;
  DoServer;
end;

var s:string;
    portnr:integer;

begin
  write('LPTDOS:parallel port server for Windows Commander. Usage: LPTDOS <portnr> (1..3)');
  write('Press any key to abort!   Copyright 2000 by Christian Ghisler, C. Ghisler & Co. ');
  write('================================================================================');
  portnr:=1;
  if paramcount=1 then begin
    s:=paramstr(1);
    if (s[1]>='1') and (s[1]<='3') then
      portnr:=ord(s[1])-ord('0');
  end;
  StartServer(portnr);
  writeln('Program aborted.');
end.

